/*
    Z80 Assembly IDE - An IDE for the Z80 line of TI calculators
    Copyright (C) 2006 Donald Straney

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <gtk/gtk.h>
#include <string.h>

#include "asmide.h"


/*
    Saves file represented
    by openfile struct [file] at
    path [path], without dialogs
    or name checking, returns
    TRUE if successful
*/

gboolean save_file_backend(struct openfile *file, gchar *path) {

    gchar *curr_text = get_file_text(file);

    if (!g_file_set_contents(path, curr_text, -1, NULL)) {	//try to save the file

		GtkWidget *dialog = gtk_message_dialog_new(NULL, 0, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, "Can't save file\n%s", path);
		gtk_message_dialog_format_secondary_text(dialog, "Are you sure you have permissions\nto write to that directory?");
		gtk_dialog_run(dialog);				//if it didn't work, show an error
		gtk_widget_destroy(dialog);
		g_free(curr_text);
		return FALSE;
    }

    g_free(curr_text);
    return TRUE;
}


/*
    Saves file represented
    by openfile struct [file],
    returns TRUE if successful
*/

gboolean save_file(struct openfile *file, struct status *curr_status) {

    gchar save_filename[MAX_FILENAME_LEN];

    if (strlen(file->filename) == 0) {		//if the file doesn't have a name yet,

		GtkWidget *dialog = gtk_file_chooser_dialog_new("Save File", NULL, GTK_FILE_CHOOSER_ACTION_SAVE, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT, NULL);
		gtk_file_chooser_set_do_overwrite_confirmation(dialog, TRUE);
		gint response = gtk_dialog_run(dialog);		//prompt for one

		if (response == GTK_RESPONSE_CANCEL) {
	    	gtk_widget_destroy(dialog);
	    	return FALSE;
		}
    
		gchar *filename = gtk_file_chooser_get_filename(dialog);	//TODO: check length!
		gtk_widget_destroy(dialog);
		strcpy(save_filename, filename);
		strcpy(file->filename, filename);	//save the file name (for if it wasn't defined at first),
		g_free(filename);
/*
	        FILE *temp = fopen(save_filename, "r");

		if (temp != NULL) {
		    fclose(temp);
	
		    GtkWidget *dialog = gtk_message_dialog_new(NULL, 0, GTK_MESSAGE_QUESTION, GTK_BUTTONS_YES_NO, "This file already exists");
		    gtk_message_dialog_format_secondary_text(dialog, "Do you want to replace it?");
		    gint result = gtk_dialog_run(dialog);
		    gtk_widget_destroy(dialog);
	
		    if (result == GTK_RESPONSE_NO)
			return FALSE;
		}
*/
    } else {
		strcpy(save_filename, file->filename);	//otherwise, use the existing name
    }

    return save_file_backend(file, save_filename);	//now try to save the file
}


/*
    Opens file [path], stores
    contents in [contents] (must
    be freed with g_free),
    length in [length], returns
    TRUE if successful
*/

gboolean open_file_backend(gchar *path, gchar **contents, guint *length) {

    if (!g_file_get_contents(path, contents, length, NULL)) {	//try to open the file

		GtkWidget *dialog = gtk_message_dialog_new_with_markup(NULL, 0, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, "Can't open file\n%s", path);
		gtk_message_dialog_format_secondary_text(dialog, "Are you sure you have permissions\nto read it?");
		gtk_dialog_run(dialog);				//if it didn't work, show an error
		gtk_widget_destroy(dialog);
		return FALSE;
    }

    if (!g_utf8_validate(*contents, *length, NULL)) {
	
		fprintf(stderr, "z80asmide: Converting %s to UTF-8...\n", path);
		guint newlength;
		gchar *newcontents = g_locale_to_utf8(*contents, *length, NULL, &newlength, NULL);
		g_free(*contents);
		*contents = newcontents;
		*length = newlength;

	//BUG: make sure this works
		if (!g_utf8_validate(newcontents, newlength, NULL)) {
		    GtkWidget *dialog = gtk_message_dialog_new(NULL, 0, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, "Can't convert this file to UTF-8");
	   		gtk_message_dialog_format_secondary_text(dialog, "It doesn't appear to be a text file...");
		    gtk_dialog_run(dialog);
		    gtk_widget_destroy(dialog);
		}
    }

    return TRUE;
}
